/*
 * This file is supposed to be included only with the Q macro already
 * defined to the value of the modulus (as a literal constant).
 * Moreover, Q is explicitly undefined at the end of this file.
 *
 * It is assumed that the file that includes this file already included
 * "ng_inner.h".
 */

#ifndef Q
#error Modulus was not defined!
#endif

/*
 * Q0I = -1/q mod 2^32
 * R = 2^32 mod q
 * R2 = 2^64 mod q
 * MSF = multiplier for mq_set()
 */
#if Q == 12289
#define Q0I   4143984639
#define R          10952
#define R2          5664
#define MSF           13
#elif Q == 18433
#define Q0I   3955247103
#define R           4564
#define R2           806
#define MSF           19
#else
#error Unknown modulus
#endif

#define Q0Ilo   ((uint32_t)Q0I & 0xFFFF)
#define Q0Ihi   ((uint32_t)Q0I >> 16)
#define Zq(name)              Zq_(Q, name)
#define Zq_(modulus, name)    Zq__(modulus, name)
#define Zq__(modulus, name)   mq ## modulus ## _ ## name

/*
 * To avoid spurious warnings about unused functions, we tag them with
 * a specific attribute.
 */
#if defined __GNUC__ || defined __clang__
#define MQ_UNUSED   __attribute__ ((unused))
#else
#define MQ_UNUSED
#endif

/*
 * IMPLEMENTATION NOTES:
 * ---------------------
 *
 * Values are kept in the [1..q] range (i.e. 0 is represented as q, not as 0).
 * Montgomery reduction of an integer x is:
 *    x <- x * Q0I
 *    x <- (x >> 16) * Q
 *    x <- (x >> 16) + 1
 * This works for all x in the [1..m] range, with:
 *    m = 2^32 - (2^16 - 1)*(q - 1)
 * The output is in the expected [1..q] range. Note that it does not work
 * for x = 0 (it returns 1 in that case, which is wrong). For the supported
 * values q (up to 18433), this implies that we can add together up to 9
 * products of reduced values, and apply one mutualized Montgomery reduction.
 *
 * When using AVX2, we keep values over 16 bits each, and can perform 16
 * Montgomery reductions in parallel.
 */

MQ_UNUSED
static inline uint32_t
Zq(add)(uint32_t x, uint32_t y)
{
	x = Q - (x + y);
	x += Q & (x >> 16);
	return Q - x;
}

MQ_UNUSED
static inline uint32_t
Zq(sub)(uint32_t x, uint32_t y)
{
	y -= x;
	y += Q & (y >> 16);
	return Q - y;
}

MQ_UNUSED
static inline uint32_t
Zq(neg)(uint32_t x)
{
	return Zq(sub)(Q, x);
}

MQ_UNUSED
static inline uint32_t
Zq(half)(uint32_t x)
{
	return (x >> 1) + (((Q + 1) >> 1) & -(x & 1));
}

MQ_UNUSED
static inline uint32_t
Zq(montyred)(uint32_t x)
{
	x *= Q0I;
	x = (x >> 16) * Q;
	return (x >> 16) + 1;
}

MQ_UNUSED
static inline uint32_t
Zq(montymul)(uint32_t x, uint32_t y)
{
	return Zq(montyred)(x * y);
}

MQ_UNUSED
static inline uint32_t
Zq(montysqr)(uint32_t x)
{
	return Zq(montyred)(x * x);
}

MQ_UNUSED
static inline uint32_t
Zq(tomonty)(uint32_t x)
{
	return Zq(montyred(x * R2));
}

/*
 * Reduce an integer x (signed) modulo q, into the [1..q] range.
 * Maximum input range for |x| is:
 *    q      max(|x|)
 *   12289    158876
 *   18433    326159
 */
MQ_UNUSED
static inline uint32_t
Zq(set)(int32_t x)
{
	return Zq(montyred)((uint32_t)(x + (int32_t)Q * MSF) * R);
}

/*
 * Same as mq_set(), but for a small input guaranteed to be lower than
 * q/2 (e.g. x is really an int8_t).
 */
MQ_UNUSED
static inline uint32_t
Zq(set_small)(int x)
{
	uint32_t y = (uint32_t)-x;
	y += Q & (y >> 16);
	return Q - y;
}

/*
 * Convert back an integer from [1..q] representation to [0..q-1]
 * representation.
 */
MQ_UNUSED
static inline uint32_t
Zq(unorm)(uint32_t x)
{
	return x & ((x - Q) >> 16);
}

/*
 * Convert back an integer from [1..q] representation to a signed
 * value in [-(q-1)/2..+(q-1)/2].
 */
MQ_UNUSED
static inline int32_t
Zq(snorm)(uint32_t x)
{
	x -= Q & (((Q >> 1) - x) >> 16);
	return *(int32_t *)&x;
}

/*
 * Division modulo q. If the divisor is not invertible (i.e. it is zero),
 * then zero is returned (represented as q).
 */
MQ_UNUSED
static uint32_t
Zq(div)(uint32_t x, uint32_t y)
{
	/* TODO: try binary GCD */
#if Q == 12289
	uint32_t y1, y2, y3, y163, y323, yt;
	y1 = Zq(tomonty(y));            /* y in Montgomery representation */
	y2 = Zq(montysqr)(y1);          /* y^2 */
	y3 = Zq(montymul)(y2, y1);      /* y^3 */
	yt = Zq(montymul)(y3, y2);      /* y^5 */
	yt = Zq(montysqr)(yt);          /* y^10 */
	yt = Zq(montysqr)(yt);          /* y^20 */
	yt = Zq(montysqr)(yt);          /* y^40 */
	yt = Zq(montysqr)(yt);          /* y^80 */
	yt = Zq(montysqr)(yt);          /* y^160 */
	y163 = Zq(montymul)(yt, y3);    /* y^163 */
	y323 = Zq(montymul)(y163, yt);  /* y^323 */
	yt = Zq(montysqr)(y323);        /* y^646 */
	yt = Zq(montysqr)(yt);          /* y^1292 */
	yt = Zq(montymul)(yt, y163);    /* y^1455 */
	yt = Zq(montysqr)(yt);          /* y^2910 */
	yt = Zq(montysqr)(yt);          /* y^5820 */
	yt = Zq(montymul)(yt, y323);    /* y^6143 */
	yt = Zq(montysqr)(yt);          /* y^12286 */
	yt = Zq(montymul)(yt, y1);      /* y^12287 = 1/y */
	return Zq(montymul)(yt, x);
#elif Q == 18433
	uint32_t y1, y7, y28, y35, y63, yt;
	y1 = Zq(tomonty(y));            /* y in Montgomery representation */
	yt = Zq(montysqr)(y1);          /* y^2 */
	yt = Zq(montymul)(yt, y1);      /* y^3 */
	yt = Zq(montysqr)(yt);          /* y^6 */
	y7 = Zq(montymul)(yt, y1);      /* y^7 */
	yt = Zq(montysqr)(y7);          /* y^14 */
	y28 = Zq(montysqr)(yt);         /* y^28 */
	y35 = Zq(montymul)(y28, y7);    /* y^35 */
	y63 = Zq(montymul)(y35, y28);   /* y^63 */
	yt = Zq(montysqr)(y63);         /* y^126 */
	yt = Zq(montysqr)(yt);          /* y^252 */
	yt = Zq(montymul)(yt, y35);     /* y^287 */
	yt = Zq(montysqr)(yt);          /* y^574 */
	yt = Zq(montysqr)(yt);          /* y^1148 */
	yt = Zq(montysqr)(yt);          /* y^2296 */
	yt = Zq(montysqr)(yt);          /* y^4592 */
	yt = Zq(montysqr)(yt);          /* y^9184 */
	yt = Zq(montysqr)(yt);          /* y^18368 */
	yt = Zq(montymul)(yt, y63);     /* y^18431 = 1/y */
	return Zq(montymul)(yt, x);
#else
#error Unknown modulus
#endif
}

/* TODO: div_x16() */

/*
 * mq_GM[x] = (2^32)*(g^rev(x)) mod q, with g a 2048-th root of 1 modulo q,
 * and rev() is the bit-reversal function over 10 bits.
 *     q      g
 *   12289    7
 *   18433   19
 */
static const uint16_t Zq(GM)[] = {
#if Q == 12289
	 10952,  11183,  10651,   1669,  12036,   5517,  11593,   9397,
	  7900,   2739,  10901,    589,    971,   1704,   4857,   5562,
	  6241,  10889,   7260,   3046,   3102,   8228,    519,   6606,
	 10000,   5956,   6332,  11479,    918,   6357,   7237,    196,
	  8614,   3587,  11068,  11665,   3165,   1074,   8124,   3246,
	  9490,  10617,    946,   1812,   2862,   6807,   6659,   7117,
	  8726,   9985,     10,   9788,   4473,   8204,  11528,   7220,
	   657,  11417,  10842,   1827,   2845,   7372,   8118,  12120,
	 11262,   7386,    672,   1521,    734,   8135,   7848,   5913,
	 12199,  10220,   8447,   4800,   6849,   8754,  12187,   3390,
	 10989,   5616,   4584,   3792,    618,   7653,   2623,   3907,
	  3775,   8270,   2759,  11676,   1514,   9681,    182,   1180,
	  2453,   9557,   9954,    256,   6264,   1450,  11792,  10012,
	   203,   6988,  12216,   9655,   5443,  11387,   9242,   8739,
	  8394,   9453,    311,   7013,   7618,   1991,  11971,   3340,
	  4457,   7290,   7841,   3977,   8601,  10525,   4232,   8262,
	  9581,  11207,  11931,   1055,   6997,  11064,    208,  11882,
	  5973,   1724,  10020,    954,   8750,  11356,  11685,   8508,
	  4350,   5786,   5458,   1491,    768,   7005,   4930,   8196,
	  9583,   8249,   1639,   9141,   4387,    219,  11680,   3614,
	 12116,  10087,   5450,   1034,   4563,  10273,   3081,   2420,
	  1684,   4031,  10170,    306,   2111,  11526,    270,   6207,
	 10670,  10435,  11721,   4420,  11376,  10826,   3900,   7730,
	  4465,   7747,   3540,  11743,  10450,   4012,    964,  12057,
	  4262,    759,   3613,   2088,   5007,   4914,   4011,   3318,
	  5112,   9376,   4397,  10007,   1767,   4164,    878,   4072,
	   106,   2983,   7529,  10732,   9138,   2798,   5855,   4200,
	  6782,   9535,    588,   2867,   9859,   5582,   6867,   6710,
	  3222,   2794,   9738,    206,  10417,   3663,  11025,   1528,
	  8132,   3703,   9062,   4601,   5436,   9451,   8397,   5016,
	    34,  11159,   9371,   2283,   4786,  12259,  10689,   6912,
	  9827,   3754,  11782,    224,   5481,   4341,  10318,   2616,
	  8221,   7251,   5761,   8047,  12181,  12264,   2763,   5760,
	  6141,  11321,   5722,   4283,  10712,   9762,   4502,   2180,
	 10873,   5134,  11648,   1786,   4530,   9924,    853,   4180,
	 10729,   9197,   3043,   9466,   8115,   4268,  10521,   9604,
	  4260,   3717,   1616,   6291,   7617,   3470,   4828,  11586,
	 10317,   4095,   9487,   2765,   5059,   1740,   6777,   4641,
	  9748,   9994,    490,    341,  10264,   8748,  11867,   9688,
	  7615,   6428,   2831,   3500,   4226,   4847,   4534,   4008,
	 11122,   5533,   8350,    795,  11388,   5367,   3593,   7090,
	  7879,   9220,   8366,   1709,   3798,  11120,   7291,   6353,
	 10034,   4826,   3414,   1473,   5704,   6327,   5637,   7108,
	   640,  11982,     12,   6830,    452,   7387,   8918,   8664,
	  9596,   1311,   8475,    255,  12000,   9605,    225,  11317,
	  9947,  10609,   8712,   6113,   8638,   4958,  10454,  10385,
	  5769,   8504,   2950,  11834,   4612,  11536,   8996,   3903,
	  9480,    829,   3250,  10538,   3623,  11876,  10744,  11590,
	  2487,   8427,   7036,   2539,  11050,   1420,  10192,   4635,
	 10030,  10742,  11709,   9879,  10924,   3439,   7271,  11355,
	  4237,    867,   9373,  11614,    765,  11442,   8079,   8356,
	  2585,  10953,   6577,   5505,   6050,  10731,   7026,   5040,
	  3812,   2703,   8981,   1510,   2385,  11817,   3501,   7979,
	  8782,    895,   6770,   2705,   5127,  11769,    941,   9207,
	  6692,   7466,   9035,   7667,   4419,   2047,   6765,  10100,
	  9872,  10933,   1414,  10113,   8201,  12253,  10369,    921,
	 12214,    324,   4991,   4000,  11852,   7295,  12204,   2825,
	  4708,   4731,   6540,  11072,    560,   7412,   6155,   2904,
	  5194,  10988,    251,   9730,   5358,   1923,   4248,   9176,
	   515,    233,   4234,   5304,   3820,   3160,   4680,   9276,
	 10410,   1727,  10180,  10094,   6584,   7441,  11798,   1138,
	  5220,   9401,   1634,   4247,   8295,   8406,   5916,      4,
	  1666,   6075,   4486,   1266,   1023,  10819,   7623,   6885,
	  2252,  11900,  12024,  10976,  10500,   3796,   1733,   5294,
	  2930,   4547,    823,  11683,  10518,   1752,   7417,   4334,
	  6144,   6884,   2573,   4123,   6797,  11928,   9421,   2067,
	  6820,   2489,   1664,   9033,   9425,   8440,   3633,   9375,
	  8555,   4825,   7457,   6619,   6426,   7632,   1503,   1372,
	 11142,    531,   3742,   7921,   9866,   7518,   7712,  10433,
	  4985,    585,   6622,    395,   7745,  10782,   9746,    663,
	 11926,   8450,     70,   7071,   6733,   8272,   6962,   1384,
	  4599,   6185,   2160,    500,   7626,   2448,   7670,  11106,
	  5100,   2546,   4704,  10647,   5138,   7789,   5780,   4524,
	 11659,  10095,   9973,   9022,  11076,  12122,  11575,  11441,
	  3189,   2445,   7510,   1966,   4326,   4415,   6072,   2771,
	  1847,   8734,   7024,   7998,  10598,   6322,   1274,   8260,
	  4882,   5454,   8233,   1792,   6981,  10150,   8810,   8639,
	  1421,  12049,  11778,   6140,   1234,   5975,   3249,  12017,
	  9602,   4726,   2177,  12224,   4170,   1648,  10063,  11091,
	  6621,   1874,   5731,   3261,  11051,  12230,   5046,   8678,
	  5622,   4715,   9783,   7385,  12112,   3714,   1456,   9440,
	  4944,  12068,   8695,   6678,  12094,   5758,   8061,  10400,
	  5872,   3635,   1339,  10437,   5376,  12168,   9932,   8216,
	  5636,   8587,  11473,   2542,   6131,   1533,   8026,    720,
	 11078,   9164,   1283,   7238,   7363,  10466,   9278,   4651,
	 11788,   3639,   9745,   2142,   2488,   6948,   1890,   6582,
	   956,  11600,   8313,   6362,   5898,   2048,   2722,   4954,
	  6677,   5073,    202,   8467,  11705,   3506,   6748,  10665,
	  5256,   5313,    713,   2327,  10471,   9820,   3499,  10937,
	 11206,   4187,   6201,   8604,     80,   4570,   6146,   3926,
	   742,   8592,   3547,   1390,   2521,   7297,   4118,   4822,
	 10607,   5300,   4116,   7780,   7568,   2207,  11202,  10103,
	 10265,   7269,   6721,   1442,  11474,   1063,   3441,  10696,
	  7768,   1343,   1989,   7629,   1185,   4712,   9623,  10534,
	   238,   4379,   4152,   3692,   8924,  12079,   1089,  11517,
	  7344,   1700,   8740,   1568,   1500,   5809,  10781,   6023,
	  8391,   1601,   3460,   7173,  11533,  12114,   7052,   3453,
	  6120,   5513,   3187,   5403,   1250,   6889,   6936,   2971,
	  2377,  11360,   7802,    213,   7132,   8023,   5971,   4682,
	  1369,   2934,   9012,   4817,   7649,   5298,  12202,   5783,
	  5242,   1441,  11312,   7170,   4163,  12001,   9218,   7368,
	 10774,   4087,   4964,   7066,  10835,  12180,  10572,   7909,
	  6791,   8513,   3430,   2387,  10403,  12080,   9335,   6371,
	  4149,   8129,   7528,  12211,   5004,   9351,   7160,   3478,
	  4120,   1864,   9294,   5565,   5982,    702,    573,    474,
	  5997,   3095,   9406,  11963,   2008,   4106,   1881,   7604,
	  8793,   9204,  11609,  10311,   3061,   7422,   2592,    600,
	  4480,  10140,     84,  10943,   3164,   2553,    981,  11492,
	  5727,   9177,  10169,   1785,  10266,   5790,   1575,   5485,
	  8184,    529,  11828,   5924,  11310,  10128,  11733,  11250,
	  3516,  10372,   8361,   9104,   7706,   7018,   1527,   2743,
	  4915,   5803,  10461,     32,    783,   9398,   1474,   7396,
	  5120,   9833,     96,   5484,   3616,   9940,   9899,   7867,
	  8765,   1460,   8229,   7708,   2734,  11784,   1741,   5751,
	  5081,   6069,   4166,   7564,   5355,   6360,   7397,   9336,
	  5806,   2937,   9172,   1668,   5483,   1383,     26,  10702,
	  2106,   6632,   1422,  10570,   4406,   8985,  12218,   6697,
	    29,   6265,  10523,   6646,  11311,   8649,   6587,   3004,
	  9977,   3106,   1800,   4513,   6355,   2040,  10488,   9255,
	  7659,   2797,   9898,   9346,   8251,  12037,  11138,   6447,
	 11764,   2268,  10359,   3422,   9230,   1909,  11694,   7486,
	  8378,   8539,   8913,   3770,   3920,   2728,   6218,   8039,
	 11780,   3182,   1757,   6665,    639,   1172,   5158,   2787,
	  3605,   1631,   5060,    261,   2162,   9831,   8182,   3487,
	 11425,  12089,   9815,   9213,   9221,   2931,   8852,   7966,
	 11962,   4362,  11438,   5151,   8909,   9686,   4545,     28,
	 11662,   5658,   6824,   8862,   7161,   1999,   4205,  11328,
	  3475,   9566,  10434,   3098,  12055,   1994,  12131,    191
#elif Q == 18433
	  4564,  17110,  12162,  16208,  10701,   9705,   3451,   5078,
	 12400,  10202,   8245,  13131,   4631,   3492,  17179,   5622,
	  5537,   3399,   2485,   9938,    345,  14064,  10152,    789,
	  5092,  15713,  12632,   6516,  16107,   2314,  15385,  17281,
	   383,   5515,   5019,  13218,   3293,   4728,   9704,  14263,
	  4417,    218,  16011,   2568,   5635,   8516,  18352,  12887,
	 13102,  15257,  14316,  12813,   8886,  11051,  13356,  15353,
	 12059,   6880,  17926,  11710,   8052,   1737,  16384,  18094,
	  3410,  14787,  13788,  14210,   2656,  17550,   7950,   4311,
	 18150,   4973,  11548,   7848,  15326,  15517,     97,  11648,
	 17990,  17685,  10847,  14695,    282,   1558,   6535,  10743,
	  2399,    181,  10165,   8051,  12204,  18401,  13377,   7233,
	 10892,  15728,  15002,  11766,   8462,  15245,  12420,   8613,
	  5053,  12360,  17415,  12678,   4606,    870,   8429,   9572,
	  6542,   1892,   4008,  17045,    371,  10155,    819,  15114,
	  1522,  13638,  16576,  17586,  16840,   7671,  13873,  12065,
	  7433,   7599,   2497,   5298,  16406,   3443,   9437,   6905,
	 14589,  17851,    209,  17496,   1698,   7028,   4444,   8211,
	  9159,  16089,  16741,   9085,   2658,   4488,   8650,   3995,
	  6532,  11903,    508,    192,   4039,  17347,  12742,   6993,
	 14812,  17645,   4527,    695,   8380,  16230,   2153,   3136,
	  2133,   4725,   9230,  13213,   6548,  18005,   6108,  16097,
	 16952,  13519,  16207,  12802,  16047,   7081,  12818,   8328,
	 17091,   8927,   9558,   9273,   9301,  10337,  11142,   5082,
	 11846,  15508,  17108,   8498,  16135,   3776,   6752,  12857,
	  3590,    486,   3056,   4203,  10364,  17125,  14532,   3025,
	 18386,  12029,   1983,   7426,  13553,    623,   6269,  15287,
	 16399,  12294,   6987,   8011,   1378,  14019,   3042,   3472,
	  4734,  12820,  16363,   7781,   3644,  16472,   3523,  14104,
	 11521,  18288,  13956,   4549,   6314,  16320,  16373,  16203,
	 15299,   7524,   9080,  15914,  17765,  12520,   5829,  13379,
	  9482,   7938,    760,  13350,   9526,  15502,  16160,   6398,
	  7067,   1655,   3428,   7827,  10564,   1235,  10800,   8291,
	 15614,  14755,   8732,   3010,  12821,   7168,   8131,   1912,
	  8093,  10461,  12301,  11616,  13947,   8029,  15138,   8334,
	 13345,  13462,   7201,  11285,   8232,   5869,   5652,   8087,
	  9232,    151,   5425,  15984,   9061,  10972,    874,   6136,
	  9299,   4966,  10442,   5398,   6605,  14398,   7625,   7091,
	 10974,  14743,   6836,  17243,    504,   7883,  10503,  12533,
	  3111,  13658,   1303,   6153,  12791,    335,  16064,   6652,
	 14432,  10970,    558,   5436,    300,  13031,  12835,   7899,
	  8435,   7252,   2970,  12879,   2786,  16438,  16584,   2204,
	  5974,   6467,   7971,  14624,   9637,   9448,  18144,   7293,
	 18121,  10042,   1398,  12430,    157,   6881,  18084,  12060,
	  5783,    444,  14853,   7936,  13337,  10411,   4401,  12549,
	 17699,   1174,   1162,   5374,   3796,    709,   1424,   8521,
	  2238,    991,   9114,  15056,   4900,  16221,    731,  18419,
	 14885,   1707,  11644,   7594,  14783,   4281,  12810,   5277,
	 10528,  15155,  16633,  13979,   5573,   7912,  15085,   4250,
	 13224,  11094,   1717,  11970,   4689,  11787,    613,  14891,
	  6892,   1734,  15910,  17044,   1022,  16497,   7473,   4421,
	 17061,   2094,  17491,  14013,   1872,  13480,  10045,  17585,
	  1562,  10460,  12143,  11266,   2168,  15769,   3047,   7683,
	 14260,   9889,  14090,  14179,   4404,  11389,  11461,   4622,
	 18349,  14047,   7466,  13272,   5005,  12487,    615,   1829,
	 13229,  15305,   3467,  11180,   2855,   8191,   3868,   9735,
	 18383,  13189,    933,   7900,  18340,  17527,   4316,  14694,
	  5680,   9549,  15669,   5777,  17938,   7070,  11080,   4478,
	  1466,  10714,  15409,   8001,   7888,   3707,  14283,   7140,
	  3046,  14214,  15419,  16423,  18200,  10217,  10615,  18381,
	 13138,   1337,   8483,   7125,   6741,  10966,  18359,   4036,
	 11656,   2954,   5907,   1652,  12095,  11393,  12093,   6022,
	 11472,   6513,  15239,  12291,  16914,   3635,   2907,    373,
	 12897,   8503,  16298,   8337,  10348,  11023,   8932,   5553,
	 12984,  11729,   9882,  13024,    556,     65,  10270,   4317,
	 14404,   9508,   9191,   9860,  14257,  11049,  13040,  14653,
	 13038,   9282,  10349,   4492,   6555,   9154,   8558,  14991,
	  4583,   3619,    379,  13206,  11105,   7100,  15820,  14978,
	  7277,  12620,   3196,  11513,   7268,  16100,     46,  12935,
	 10191,   4142,   9281,  11926,  14900,  14340,  16894,   5224,
	  9309,  13388,  13942,   3818,   2937,   7206,  14135,  15212,
	  7925,   1689,   8800,   1294,   5524,  14570,  16368,  11992,
	  9491,   4458,   3910,  11928,  13598,   1656,   3586,   8177,
	 13056,   2322,  16649,   1648,  14699,  18328,   1843,    116,
	 10016,   4221,   3330,   2710,   5358,  11169,  13567,   1354,
	  8715,   3439,   8805,   5505,  10680,  17825,  14534,   8396,
	  4185,   3904,   8543,   2358,  13314,  13160,  14784,  16183,
	  3842,  13644,  17524,   1253,  13782,  16530,  12687,  15971,
	 13700,  17515,   2420,  10494,   7049,   8615,  15561,  10671,
	 10485,   1060,   1583,   2340,   6599,  16718,   5525,   8039,
	 12196,  15350,  10577,   8497,  16786,  10118,  13406,   2164,
	   696,   7375,   3971,    630,  13829,   4501,  10704,   8545,
	  8124,  10763,   4718,   6718,  13636,  11540,  16886,   2173,
	 13510,   4961,   9652,   3648,   3009,  16232,   2469,   3836,
	  4933,   3461,  12281,  13205,  11756,  13442,   4041,   4285,
	  3661,  16043,   9473,  11418,  13814,  10301,   5454,  10915,
	  8727,  17232,  13005,   3609,   9965,   5508,   3913,  10768,
	 11368,   3716,  15705,  10290,  10822,  12073,   8935,   4393,
	  3878,  18157,  11691,  13998,  11637,  16445,  17690,   4654,
	 12911,   9234,   2765,   6125,  12586,  12014,  18046,   2176,
	 17540,   7355,    811,  12063,  17878,  11837,   8513,  13958,
	 16653,  12390,   3722,   4745,   7749,   8299,   2499,  10669,
	 16214,   3951,  15969,    375,  13937,  18040,  11638,   9914,
	 16136,  15678,   7102,  12699,   9368,  15152,  16159,  12929,
	 14186,  13925,   6623,   7438,   5741,  16684,    153,  14572,
	 14261,   3358,  14440,  14021,  15097,  18043,  12112,  10964,
	  5242,  13012,   9833,   1249,  16386,   5032,   2437,  10065,
	  1738,   3850,     11,   1891,   3970,   7161,   7025,  17895,
	  6303,  14429,  12523,  17941,   6931,   5087,  11127,  10882,
	 13926,  16149,   7788,  11652,   8944,    913,  15223,   6189,
	  9511,   2869,  10910,   8768,   6262,   5705,  16606,   5986,
	 10784,   2189,  14068,  10397,  14897,  15500,  15844,   5698,
	  5743,   3622,    853,  14256,   9576,   2313,  15227,  16931,
	  3810,   1440,   6324,   6309,   3400,   6365,  10288,  15790,
	 16146,   5667,  10602,  11119,   5700,   7960,   4236,   2617,
	 12801,   8757,   1131,   5072,  16068,  17394,   1735,   5010,
	  2908,  12275,   3985,   1361,  17206,  13615,  12942,   9536,
	 12505,   6468,   8129,  14974,   2983,   1708,  11802,   7944,
	 17712,   8436,   5712,   3320,  13774,  13479,   9887,  17235,
	  4487,   3873,   3645,   9941,  16825,  13471,   8623,  14435,
	  5656,    396,   7269,   9569,    935,  13271,  13889,  18167,
	  6320,  14000,     40,  15255,   4382,   7607,   3761,   8098,
	 15702,  11450,   2666,   7539,  13722,   2864,  10120,   7018,
	 11627,   8023,  14190,   6234,  15359,   2757,  11647,   6434,
	  1917,  14513,   7362,  10475,    985,     82,  12956,  10267,
	 10798,   2920,    535,   8185,  17135,  16491,   6525,   2321,
	 11245,  14410,   9521,  11291,   4326,   4683,   2594,  16946,
	 12878,   3561,   9648,  11339,   9944,  13628,  14996,  14086,
	 16837,   8831,  12823,  12539,   2930,  16057,  11685,  16318,
	 11722,  14300,  10574,   9657,  17379,   8165,  18193,    635,
	 17483,  10962,  17727,   2636,  16666,   1219,   8272,   2691,
	 15755,  15534,   2783,  17598,   9028,   5299,   7757,  11350,
	  9421,    803,  16276,   4555,   2408,  15134,  13315,   6629,
	  2575,  12004,  16466,  17109,  14006,   9793,  17355,  17445,
	  9993,   6970,  13713,   6344,  17481,   5591,  17027,   2952,
	   268,    827,   1635,  12955,   8609,  13704,   8571,   3820,
	 15205,  13149,  13046,  12333,   8005,  13766,  18367,   7087,
	  5414,  14093,  14734,  10939,  12282,   6674,   3811,  13342
#else
#error Unknown modulus
#endif
};

/*
 * mq_iGM[x] = (2^31)*((1/g)^rev(x)) mod q.
 * (Note: 2^31 instead of 2^32 because we want to pre-divide it by 2)
 */
static const uint16_t Zq(iGM)[] = {
#if Q == 12289
	  5476,    553,   5310,    819,   1446,    348,   3386,   6271,
	  9508,   3716,  11437,   5659,   5850,    694,   4775,   8339,
	 12191,   2526,   2966,  11830,    405,   9123,   9311,   7289,
	  8986,   5885,   8175,  10738,  10766,   8659,    700,   3024,
	  6229,   8230,   8603,   4722,   5231,   6868,    436,   5816,
	  8679,   6525,   8187,   3908,   7395,  12284,   1152,   7926,
	  2586,   2815,   2741,  10858,  11383,  11816,    836,   7544,
	 10666,   8227,  11752,   4562,    312,   6755,   4351,   7982,
	  8158,  10173,    882,   1844,   4156,   2224,   8644,   3916,
	 10619,    159,   5149,   8480,   2638,   5989,   1418,   8092,
	  1775,   7668,    451,   3423,   1317,   6181,   8795,   6043,
	  7283,   6393,  11564,   9157,  12161,   7312,   1366,   4918,
	 11699,  12198,   1304,  11532,   6451,   4765,   8154,   4257,
	  4191,   4833,   2318,  11980,  10393,   9997,   9481,    650,
	 10594,     51,   7912,   2720,   9889,   1921,   7179,     45,
	  3188,   8365,   2077,  11922,   5384,  11953,   8596,   6658,
	 10981,   7130,   3974,   3404,  12177,   6398,  10412,   1231,
	  8833,    800,     15,   9896,   5003,   1459,    565,  12272,
	  9781,   1946,   1419,   9571,   3844,   7758,   4293,   8223,
	 11525,    632,   4313,    936,  12186,   7420,  10892,  10678,
	  8934,   2711,   9498,   1215,   4711,  11995,   1377,   8898,
	 10189,   3217,  10890,   7720,   6923,   2380,   4653,  12236,
	 10253,  11850,  10207,   5261,   1141,   3946,   7601,   9733,
	 10630,   4139,   9832,   3641,  11245,   4338,   5765,  10158,
	   116,  11807,  10283,   7064,    273,  10519,   2271,   3912,
	  8424,  10339,   6876,   6601,  10079,    284,    927,   6954,
	  3041,  12154,   6526,   5089,  12136,   7204,   4129,  11447,
	 11079,   4604,   1008,   3863,  11772,   9564,   1101,   6231,
	 10482,   6449,   6035,   3951,   1574,   5325,   2020,   1353,
	  8191,   9824,   2642,  11905,   5399,   9560,   9396,  10114,
	  8035,    302,   6611,   7914,  11812,   7279,  11427,   3158,
	  6348,  12185,   6757,   2646,   5617,    179,    541,   1354,
	  9642,   5278,  10391,   7039,   6801,   6277,   6339,  11163,
	  2702,   2333,    735,   5633,  11656,  10046,   3107,  11456,
	 12287,   9331,   8086,   1997,   4021,  11472,   1444,   9679,
	 11720,   6390,   2424,   8997,   7242,   7199,   5281,   7084,
	  7651,   9949,  10709,  10379,   9637,  10172,   6028,   5887,
	  7701,  10165,   5183,   9610,   7424,   6019,   6795,   9692,
	 10837,   3067,   8583,  12009,   6753,   9019,   3779,   9935,
	  4732,   6187,   2497,   6363,  10289,   3649,  12127,   6182,
	  5684,    960,     18,   2044,   1088,  11582,    678,   7353,
	  7239,   2762,   5121,   3935,   2311,   1627,   8556,   8943,
	  1541,   5674,    260,   3581,   4792,   8904,   5697,   7898,
	  2155,   4394,    236,   4952,  11534,   1654,   4793,  10383,
	  9769,   8776,    779,   9264,   3392,   2856,    668,   4852,
	  8111,   2105,   6568,   5762,   6482,   1458,   5711,   4026,
	   467,   2509,   4425,   6827,   1205,    290,   6918,   7274,
	  3827,   7193,  11579,   6764,   4875,   8771,   1931,   4901,
	  6494,   6917,   6351,   4333,   7020,  10664,   5730,   7549,
	  4193,   7791,   6521,   9983,   6372,  10814,   8037,   3260,
	   952,   7062,   9810,   7970,   3088,   7933,    840,   1171,
	   486,   6032,   1342,   6289,   6017,   1907,   5489,   7491,
	  7957,   7830,   2451,  12063,   8874,  12283,   6298,  11969,
	  8735,   3326,   2981,   9437,   5408,  10582,   9876,   7272,
	  2968,   2499,   6729,  10390,   5290,   8106,   7679,   2205,
	  8744,   4348,   3461,   6595,   5747,   8114,   3378,   6728,
	 10285,  10022,   3721,  10176,  10539,   4729,   9075,   2337,
	  7445,    211,   7915,   7157,   5974,  12044,   7292,   7415,
	  3824,   2756,  11419,   3615,   4762,   1401,   4097,    986,
	  6496,   9875,  10554,   2336,   2999,  11481,   4286,  10159,
	  7487,    884,  10155,   2087,   7556,   4623,   1546,    780,
	 10199,   5718,   7327,  10024,  11396,   6465,   9722,    708,
	 11199,  10038,   7408,   6933,   4003,   9428,    484,   3074,
	  9409,   4763,   6157,     54,   2121,   3264,   2519,   2034,
	  6049,     79,  11292,    117,  10740,   7072,   7506,   4407,
	  6625,   4042,   5145,   2564,   7858,   8877,   9460,   6458,
	 12275,   3872,   7446,   1690,   3569,   6570,  10108,   6308,
	  8306,   7863,   4679,   1534,   1538,   1237,    100,    432,
	  4401,   8198,   1229,  11208,   6014,   9759,   5329,   4342,
	  4751,   9710,  11703,   5825,   2812,   5266,  10698,   6399,
	  2125,   9180,  10925,  10329,  10404,   1688,   1875,   8100,
	  8546,   6442,   5190,   7674,  10578,    965,  11155,   6407,
	  2921,   6720,    126,   2019,   7616,   7340,   4746,   2315,
	  1517,   7045,  11269,   2967,   3888,  11389,  10736,   1156,
	 10787,   2851,   1820,    489,   8966,    883,   3012,   6130,
	  2796,   6180,   1652,  10086,   7004,  11578,   8973,  11236,
	  6938,  12276,   5453,   3403,  11455,   7703,   4676,   9386,
	  7621,   2446,   9109,   3467,   8507,  10206,   3110,   3604,
	  3269,   5274,   6397,  10922,   8435,   2030,  11559,   1762,
	  2211,   1195,   7319,  10481,   9547,  12241,   1228,   9729,
	  8591,  11552,   7590,   5753,  12273,    914,   3243,   3687,
	  4773,   5381,   8780,   8436,   7737,   1964,   7103,  10531,
	  6664,    278,   7225,   6634,   9327,   6375,   5880,   8197,
	  3402,   5357,   9394,   7156,   5252,   1060,   1556,   3281,
	  6543,   5654,   4868,  10707,    673,  12247,   7219,  10049,
	 11989,  10993,   8578,   4614,    989,    340,   7687,   1748,
	  8487,   5204,  10236,  11285,    163,   7586,   4597,   3146,
	 12052,   5858,  11938,   9298,   3362,   7642,  11357,  10229,
	 10550,   8709,   1469,   9787,     39,   8525,   2080,   4070,
	  2959,   1477,   6249,    943,   4951,  10574,   1888,   2749,
	  2190,   7003,   6199,    727,   8756,   9807,   4101,   6902,
	  8605,   7680,    144,   4063,   8704,   6633,   5424,   9668,
	  3253,   6188,   9640,   2320,   3736,   7783,  10822,   5460,
	  9948,   3159,   2133,   8723,   6038,   8388,   6609,   4956,
	  4659,   8821,   2700,  11664,   3443,   4551,   3388,   9229,
	  4418,   8763,   6232,    378,   2558,  10559,   5344,   1949,
	  3133,    754,   3240,  11539,  11505,   7919,  11439,   8617,
	   386,   5600,    105,   7827,  10443,  10213,   3955,  12170,
	  7022,   1333,   9933,   5552,   2330,   5150,   5473,   8405,
	  6941,   4424,   5613,   6552,  11568,   2784,   2510,   1012,
	  1093,   6688,   5041,   8505,   8399,  10231,   9639,    841,
	  9878,  10230,   2496,   4884,  11594,   4371,   7993,  11918,
	 10326,   9216,  10004,  12249,   7987,   3044,   4051,   6686,
	   676,   4395,   7379,    909,   4981,   5788,   3488,   9661,
	   812,   8915,  10536,    292,   1911,  12188,   3608,   2806,
	  9812,  10928,  11265,   9340,   9108,   1988,   6489,  11811,
	  8998,  11344,   8815,  11045,  11218,   1272,   4325,   6395,
	  3819,   7650,   7056,   2463,   8670,   5503,   7707,   6750,
	 11929,   8276,   5378,   3079,  11018,    408,   1851,   9471,
	  8181,   7323,   6205,   9601,    926,   5475,   4327,   9353,
	  7089,   2114,   9410,   6242,   8950,   1797,   6255,   9817,
	  7569,  11561,  10432,   6233,   2452,   1253,   3787,   9478,
	  7950,   9766,   6174,    619,   4514,   3279,  11352,   2834,
	   599,   1113,  11465,  10204,   6177,   5056,   9926,   7488,
	   136,   4520,   3157,  11672,   9219,   6400,    120,   5434,
	  1825,   7884,   7214,   2654,  11393,   2028,   9562,   9848,
	  8159,  11652,   9128,   6990,   8290,   8777,   7922,   5221,
	  4759,   9253,   3937,  10126,  11306,   8534,   4922,   4550,
	   424,    357,   6228,   6751,   7778,   1158,   1097,    315,
	 10027,   9399,   2250,   9720,    821,   9937,  11016,   9739,
	  6736,   8454,  11065,   8476,  12039,  11209,   3052,   3845,
	 11597,   8808,   8153,   2778,   2609,  12254,   8064,   6326,
	  5813,   7416,   6898,   2272,   5947,   8978,   5852,   3652,
	   928,   8433,   8530,   7356,   2184,  10418,   5879,   6718,
	 11603,   5393,   8473,   9076,   2835,   2416,   3732,   1867,
	  1457,   4328,   8069,   1432,   1628,  11457,   4900,   8879,
	  5111,   1434,   6325,   2746,   4083,   4858,   8847,   9217,
	 10122,   2436,  11413,   7030,    303,   5733,   3871,  10824
#elif Q == 18433
	  2282,   9878,  10329,  12352,  15894,   7491,   4364,   3866,
	 15622,    627,  16687,   6901,   2651,   5094,  13332,  12233,
	   576,   1524,  17276,   1163,  15175,  12117,   1360,  15887,
	  8822,  13357,  11401,   9044,  13464,   7974,   7517,   6448,
	  9386,  10241,   8348,  14407,  12578,   9470,  14993,   3187,
	  1540,  11755,   3691,  13990,   2810,  11275,   1588,  11882,
	  2773,   9257,  14175,   6399,  17149,   1211,  18324,   7008,
	  2085,  13581,  16069,   7570,  11824,   6707,   6459,   9025,
	  3184,   2280,   5381,  10013,   9640,  10145,  11614,  17672,
	 10876,   8807,   4139,   9031,    694,  16429,  17487,  15162,
	 13647,   5002,  17998,  16130,  12094,    509,  12253,   6690,
	  4910,  12223,   1594,  14202,  12550,  10932,  10569,  12987,
	  5600,   2528,     16,  12331,   5191,   4134,   9126,   8017,
	  3845,   5949,  17654,  18292,   1869,   3793,    374,   9438,
	 12609,   9168,   1458,  10770,  14509,  12659,   6730,   9358,
	  7061,  14458,   9658,  17105,  11328,  11539,   1823,  16728,
	 15234,  10353,  10682,  13670,  11758,  18053,  14464,  13692,
	  2527,   6302,  12173,    334,  10476,  13893,  14671,   1567,
	  1115,   1030,  10273,  15276,   6942,  11455,   9289,   3456,
	 11381,   7455,  10197,  16611,   5326,   1035,  12023,  16066,
	 16697,  16912,   2207,  17744,   5211,   5723,  12286,   1017,
	  1573,   6082,   8905,   2440,  14720,   8225,   3202,   9240,
	  7704,  11167,    654,  13251,   7115,  16905,  18190,  16638,
	  2788,  15057,  16545,   1149,  14184,   9879,  10679,  12510,
	 15892,  12862,   4048,   4566,   4580,  13654,   4753,    671,
	 14269,  12024,   5676,   1193,  12032,   1113,   2457,   9957,
	  1168,  15379,    214,  15159,   2610,  13818,   6854,   8150,
	 16865,   8140,  10318,  14243,   8869,   6953,    394,  11027,
	  5720,  12062,    543,   7197,  18337,  18179,   3265,  15167,
	  7219,  14108,  16189,  17104,   4674,    846,   1172,   4637,
	  5111,  16211,  14919,  17584,   9685,   9112,    291,   1922,
	  5764,   4498,   7495,  10230,  15784,   7968,   5417,   5500,
	  6440,  13967,   3705,  13259,   5048,  10284,   4965,   2768,
	  9030,   7763,   7399,   9976,   3071,   1597,   5960,  12697,
	 15422,   3170,   3520,   3169,  17607,   6263,  16956,  12605,
	 16415,     37,  12950,   5846,   5654,   4975,   8548,  11864,
	    26,   3909,   4108,   9333,   1005,   1507,  11326,  16910,
	 14863,   2075,   7363,  14489,   5216,   1512,  13076,  17700,
	 16194,  12893,  14898,   9464,   6328,   1382,   4442,  15593,
	 11086,  16275,    453,   9263,  14483,   8750,   2622,     25,
	  4349,  16499,   5121,   7789,  12843,   7483,   1564,   2602,
	  8302,   8909,   2973,   6714,  11797,  14700,   2193,     42,
	 16122,   3486,   3522,  16231,   2127,  11388,   4272,  11303,
	  5375,   7693,   1332,  17349,  12800,   3145,  13203,  17652,
	   424,   4194,  11693,  17497,   2210,    471,  17386,    686,
	  7006,   5480,    968,  17922,   9911,  10478,  17566,  14987,
	  1771,   8910,   3323,   6872,  12448,   8358,  12886,  11821,
	 16308,   1674,  14477,   6430,   2227,    900,   1639,  13169,
	  6578,  12028,   7076,   1825,  14636,  12611,   8363,   1774,
	     7,   8851,   1106,  15983,  10905,  13876,   8721,  17314,
	  4956,  17721,   8862,  16535,  15746,  17852,  17846,    367,
	  2942,   7016,   4011,   2548,  14465,   1790,  18211,   6325,
	 12403,   9391,   5776,   9138,  12218,  17734,  13412,    156,
	  5570,   9361,  13709,   4398,  11121,   5231,   5983,  15446,
	 17331,  10141,  10214,  17040,   2777,  16948,  14807,   4999,
	  5267,   2799,   2701,  18283,  15715,  18154,  12948,  11217,
	 15107,  10401,   9049,   2821,   6140,   8565,  11604,   7661,
	  2950,   3965,   5275,  18181,    595,  15015,   1845,  12946,
	  5671,   5404,  11234,   5914,  15734,  13212,  15950,   4567,
	 15365,  17996,  12947,   4686,  10441,   6504,   9141,  13817,
	  5173,  15607,   6282,  14317,   3574,   5616,  11702,   2544,
	 14266,  10864,   5202,   2243,  12625,   3066,   3986,   5170,
	 17477,   5151,  14849,   2806,  16928,  14067,   1839,  10626,
	  5071,  13033,   8599,  13151,   5303,  16719,   8389,   5683,
	 11762,   7311,  15096,  12292,   3747,  11066,   2170,  15726,
	  5673,     33,  11550,   5214,   3050,  11910,   2642,   1614,
	 16523,   4931,  11581,   4912,   2739,   8399,   8803,  18299,
	 16957,    703,   6421,    476,  15261,   2360,  14948,   4220,
	   494,    539,   4320,  11430,    662,  10200,  12431,   7929,
	  5902,   2559,  10866,  17229,   6939,  10295,   8815,   4506,
	 12758,   5338,   6567,  13919,   9634,   7825,  10666,   1339,
	  7871,  14297,   8607,  10100,  17115,    353,  12952,    475,
	  8899,    120,   5134,    527,   4388,  13146,  11283,  12572,
	 10274,   3374,   1188,  16968,   2947,   2805,   4801,    798,
	 11390,  10935,  11619,  13461,   3547,  13609,   7436,  11994,
	  9960,  17136,   6875,  16270,   3571,   4456,  11228,   3594,
	  8056,   5954,    971,    649,   5124,   8949,  16973,  13034,
	  4083,  11955,  18392,   8724,   3979,  14752,   1960,   8258,
	 15216,   3393,   7838,   1537,  15316,  11338,   5205,   3403,
	 14924,  13373,  17001,  11572,   5447,  17100,  12708,  10582,
	 14384,   7336,   5413,  16242,   1589,  18413,  11433,  15273,
	   133,   2272,   2581,   8749,   4432,   5582,  18235,  15605,
	  1999,   4905,   2481,    804,   4246,   7394,   7280,   6973,
	   599,   4273,   2477,  11546,  16773,  15577,  14215,   9577,
	 14461,  12532,  17579,   7725,  10946,   5152,  15199,   2964,
	 13665,  11962,   2409,   9830,   8536,   7224,   3079,  16979,
	 15928,   8349,   9736,  10399,  15897,   8651,   4838,   2816,
	  7908,  16315,  14453,  15583,   3657,  13132,   6383,  10360,
	 10538,  13289,   6034,  16733,   6062,  15271,  17713,  16528,
	   751,   1603,   8060,  13645,  11305,   8790,  16622,   6345,
	 15584,  10511,  10683,   1768,   4018,  11399,   8122,  13041,
	 15440,  10130,   6364,  15302,  14049,  12978,   7782,   4461,
	  6122,   1605,   8760,  13961,  12607,  14539,   1142,  11470,
	 12992,   3653,   6673,   5751,    246,   2955,   2002,   6065,
	   269,   5704,   5636,  16448,   8271,   9211,  16508,  17564,
	  4184,   7998,  15917,  10240,   8592,   4300,  11927,  15812,
	 12951,  12377,    195,   1668,   2206,  11213,  16754,   2086,
	 11147,   9140,  10091,   6346,  14714,   5905,   2254,  11340,
	  2752,   1137,  10857,  13749,   2867,  14882,  10594,  10365,
	 13476,  12614,   9413,   2248,   9029,   1232,   7241,  10326,
	  3882,   7967,   5067,   5342,   6844,  16572,  12238,    890,
	 11454,   4960,   3298,   9494,   3185,   8811,   5539,   9663,
	 17345,   9410,  12426,  12140,   6154,   7834,  13816,   2761,
	 16106,   9588,    994,   3398,  11434,   3371,    138,  16494,
	  7020,   4749,   3180,  13022,  13288,   1364,  16575,  12749,
	 13049,   7260,  15679,   4234,   7412,   2714,   9817,   4853,
	  3759,  15706,   4066,  11526,  12724,   4480,   1195,   7386,
	  7074,   7196,  11712,  12555,   2614,   3076,   7486,   6750,
	 16515,   7982,  10317,   7712,  16609,  13607,   6736,  11678,
	  8130,   9990,  12663,  11615,  15074,  16074,   3835,  14371,
	  4944,  13081,   6966,   2302,  18118,   7231,   5529,  18085,
	 17351,  11730,  13374,  10040,   4968,   3928,  10758,  12335,
	  5197,   6454,  10074,   5917,  17263,   8425,  17903,   3974,
	  3881,   1436,   4909,   5692,  13186,  17223,    459,  11583,
	  1231,   2873,  10168,  11542,   8590,   9671,  11611,  16512,
	  1125,  11041,  11853,  11776,  17254,   4945,  16481,   7124,
	 14235,  11166,    304,  13093,   6464,   4814,   7497,   4859,
	 17756,   2433,   3632,  15754,  17078,  16768,   7106,  13425,
	 18375,   8295,   9269,   1867,  17609,    892,  17272,  11905,
	  5128,  16640,  17605,  11634,  12469,  16478,  16204,   4471,
	 12437,  10249,  11148,  15671,  17786,  14033,   8372,   5254,
	 10827,   2149,  14830,   7748,  16524,  11462,  11739,   4562,
	 15821,   9986,  11263,  10983,  12470,   4576,  16362,   4121,
	  2749,  18410,  10383,  14799,   3460,  16835,  12123,   5578,
	 10944,  10523,  14883,   3664,  11830,   9027,   7407,   6925,
	  1721,  14154,  13856,   5939,  16187,   4042,  13792,  11914,
	  1890,  11913,   3692,   2088,  13503,   4621,  13679,  11231,
	  7058,  13298,   9184,  18155,  11921,  13492,   3352,  11941
#else
#error Unknown modulus
#endif
};

MQ_UNUSED
static void
Zq(NTT)(unsigned logn, uint16_t *restrict a)
{
	size_t t = (size_t)1 << logn;
	for (unsigned lm = 0; lm < logn; lm ++) {
		size_t m = (size_t)1 << lm;
		size_t ht = t >> 1;
		size_t v0 = 0;
		for (size_t u = 0; u < m; u ++) {
			uint32_t s = Zq(GM)[u + m];
			for (size_t v = 0; v < ht; v ++) {
				size_t k1 = v0 + v;
				size_t k2 = k1 + ht;
				uint32_t x1 = a[k1];
				uint32_t x2 = Zq(montymul)(a[k2], s);
				a[k1] = Zq(add)(x1, x2);
				a[k2] = Zq(sub)(x1, x2);
			}
			v0 += t;
		}
		t = ht;
	}
}

MQ_UNUSED
static void
Zq(iNTT)(unsigned logn, uint16_t *restrict a)
{
	size_t t = 1;
	for (unsigned lm = 0; lm < logn; lm ++) {
		size_t hm = (size_t)1 << (logn - 1 - lm);
		size_t dt = t << 1;
		size_t v0 = 0;
		for (size_t u = 0; u < hm; u ++) {
			uint32_t s = Zq(iGM)[u + hm];
			for (size_t v = 0; v < t; v ++) {
				size_t k1 = v0 + v;
				size_t k2 = k1 + t;
				uint32_t x1 = a[k1];
				uint32_t x2 = a[k2];
				a[k1] = Zq(half)(Zq(add)(x1, x2));
				a[k2] = Zq(montymul)(s, Zq(sub)(x1, x2));
			}
			v0 += dt;
		}
		t = dt;
	}
}

MQ_UNUSED
static inline void
Zq(poly_set_small)(unsigned logn, uint16_t *d, const int8_t *a)
{
	size_t n = (size_t)1 << logn;
	for (size_t u = 0; u < n; u ++) {
		d[u] = Zq(set_small)(a[u]);
	}
}

/*
 * Convert a small polynomial (signed 8-bit coefficients) to a mod q
 * representation that _starts_ at the same address (i.e. the n first
 * bytes of d are read, and 2*n bytes are written into d).
 */
MQ_UNUSED
static inline void
Zq(poly_set_small_inplace_low)(unsigned logn, uint16_t *d)
{
	size_t n = (size_t)1 << logn;
	size_t u = n;
	while (u > 0) {
		u -= 2;
		uint32_t x = d[u >> 1];
		uint8_t x0 = (uint8_t)x;
		uint8_t x1 = (uint8_t)(x >> 8);
		d[u + 0] = Zq(set_small)(*(int8_t *)&x0);
		d[u + 1] = Zq(set_small)(*(int8_t *)&x1);
	}
}

/*
 * Convert a small polynomial (signed 8-bit coefficients) to a mod q
 * representation that _ends_ at the same address (i.e. the n last
 * bytes of d are read, and 2*n bytes are written into d).
 */
MQ_UNUSED
static inline void
Zq(poly_set_small_inplace_high)(unsigned logn, uint16_t *d)
{
	size_t n = (size_t)1 << logn;
	size_t hn = n >> 1;
	for (size_t u = 0; u < n; u += 2) {
		uint32_t x = d[hn + (u >> 1)];
		uint8_t x0 = (uint8_t)x;
		uint8_t x1 = (uint8_t)(x >> 8);
		d[u + 0] = Zq(set_small)(*(int8_t *)&x0);
		d[u + 1] = Zq(set_small)(*(int8_t *)&x1);
	}
}

MQ_UNUSED
static inline void
Zq(poly_snorm)(unsigned logn, uint16_t *d)
{
	size_t n = (size_t)1 << logn;
	for (size_t u = 0; u < n; u ++) {
		d[u] = (uint16_t)Zq(snorm)(d[u]);
	}
}

#undef Q
#undef Q0I
#undef R
#undef R2
#undef MSF
#undef Q0Ilo
#undef Q0Ihi
#undef Zq
#undef MQ_UNUSED
